"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var htmlparser2_1 = require("htmlparser2");
var Factory_1 = require("./parser/Factory");
var FcAttributeHandler_1 = require("./parser/FcAttributeHandler");
var FcCallTagHandler_1 = require("./parser/FcCallTagHandler");
var FcContentAttributeHandler_1 = require("./parser/FcContentAttributeHandler");
var FcContentTagHandler_1 = require("./parser/FcContentTagHandler");
var FcEachTagHandler_1 = require("./parser/FcEachTagHandler");
var FcElseIfTagHandler_1 = require("./parser/FcElseIfTagHandler");
var FcElseTagHandler_1 = require("./parser/FcElseTagHandler");
var FcIfTagHandler_1 = require("./parser/FcIfTagHandler");
var FcKeyAttributeHandler_1 = require("./parser/FcKeyAttributeHandler");
var FcTagHandler_1 = require("./parser/FcTagHandler");
var utils_1 = require("./parser/utils");
var DEFAULT_TAG_HANDLERS = {
    '*': new FcTagHandler_1.FcTagHandler(),
    'fc-each': new FcEachTagHandler_1.FcEachTagHandler(),
    'fc-if': new FcIfTagHandler_1.FcIfTagHandler(),
    'fc-else': new FcElseTagHandler_1.FcElseTagHandler(),
    'fc-else-if': new FcElseIfTagHandler_1.FcElseIfTagHandler(),
    'fc-content': new FcContentTagHandler_1.FcContentTagHandler(),
    'fc-call': new FcCallTagHandler_1.FcCallTagHandler()
};
var DEFAULT_ATT_HANDLERS = {
    '*': new FcAttributeHandler_1.FcAttHandler(),
    'fc-key': new FcKeyAttributeHandler_1.FcKeyAttHandler(),
    'fc-content': new FcContentAttributeHandler_1.FcContentAttHandler()
};
var DEFAULT_OPTIONS = {
    output: 'function',
    pretty: false,
    interpolation: /\{\{([\s\S]+?)\}\}/gm,
    propNamePrefix: '#',
    elVarName: 'el',
    ctxVarName: 'ctx',
    selfClosingElements: [
        'area', 'base', 'br', 'col', 'embed', 'hr', 'img', 'input',
        'keygen', 'link', 'meta', 'param', 'source', 'track', 'wbr'
    ],
    tagHandlers: DEFAULT_TAG_HANDLERS,
    attHandlers: DEFAULT_ATT_HANDLERS
};
/**
 * Parse an HTML template and return the factory of the underlying render function.
 * @param {string} html the HTML content to parse
 * @param {ParserOptions} options the options
 * @return {string | RenderFactory} the factory function as Function or string.
 */
function parse(html, options) {
    if (options === void 0) { options = {}; }
    options = utils_1.assign({}, DEFAULT_OPTIONS, options);
    options.tagHandlers = utils_1.assign({}, DEFAULT_TAG_HANDLERS, options.tagHandlers);
    options.attHandlers = utils_1.assign({}, DEFAULT_ATT_HANDLERS, options.attHandlers);
    var factory = new Factory_1.Factory(options);
    var p = new htmlparser2_1.Parser({
        onopentag: function (name, attrs) {
            var tagHandler = options.tagHandlers[name] ? options.tagHandlers[name] : options.tagHandlers['*'];
            if (tagHandler) {
                tagHandler.startTag(factory, name, attrs, options);
            }
        },
        onclosetag: function (name) {
            var tagHandler = options.tagHandlers[name] ? options.tagHandlers[name] : options.tagHandlers['*'];
            if (tagHandler) {
                tagHandler.endTag(factory, name, options);
            }
        },
        ontext: function (text) {
            factory.appendText(utils_1.interpolate(text, options));
        },
        oncomment: function (text) {
            factory.appendComment(utils_1.interpolate(text, options));
        }
    }, {
        xmlMode: false,
        decodeEntities: false,
        lowerCaseTags: true,
        lowerCaseAttributeNames: true,
        recognizeSelfClosing: true,
        recognizeCDATA: true
    });
    p.parseComplete(html);
    return factory.toFunction();
}
exports.parse = parse;
