"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var utils_1 = require("./runtime/utils");
/**
 * The current root element, i.e. where the starting point of the updateElement function.
 * @type {Element}
 */
var rootElement = null;
/**
 * The current document owning the current root element
 * @type {Document}
 */
var rootDocument = null;
/**
 * The current parent element of the processed instruction.
 * @type {Element}
 */
var parentElement = null;
/**
 * Build a context object.
 * @return {Context} the context
 */
function getCtx() {
    return { root: rootElement, parent: parentElement, document: rootDocument };
}
/**
 * Restore a context object.
 * @param {Context} ctx the context to restore
 */
function restoreCtx(ctx) {
    rootElement = ctx.root;
    parentElement = ctx.parent;
    rootDocument = ctx.document;
}
/**
 * Remove the un-visited nodes.
 */
function cleanRemainingNodes() {
    if (rootElement.__content__ !== parentElement) {
        while (parentElement.childNodes.length > lastIndex()) {
            parentElement.removeChild(parentElement.lastChild);
        }
    }
}
/**
 * Get and increment the fcIndex property of the current parent.
 * @return the index
 */
function nextIndex() {
    if (!parentElement.fcIndex) {
        parentElement.fcIndex = 0;
    }
    var index = parentElement.fcIndex;
    parentElement.fcIndex = parentElement.fcIndex + 1;
    return index;
}
/**
 * Get fcIndex property of the current parent.
 * @return the index
 */
function lastIndex() {
    return parentElement.fcIndex || 0;
}
/**
 * Clear the fcIndex property of the current parent.
 */
function clearIndex() {
    delete parentElement.fcIndex;
}
/**
 * Handle a Node.
 * @param {string} value the node's value
 * @param {number} nodeType the node's type
 * @param {NodeFactory} factory the factory creating the node when necessary
 * @return {Node} the handled node
 */
function handleNode(value, nodeType, factory) {
    var index = nextIndex();
    var current = parentElement.childNodes.item(index);
    if (current) {
        if (current.nodeType !== nodeType) {
            parentElement.insertBefore(factory(value), current);
        }
        else {
            current.nodeValue = value;
        }
    }
    else {
        parentElement.appendChild(factory(value));
    }
    return current;
}
/**
 * Create an element
 * @param {string} name the name
 * @param {string[]} attrs the attributes
 * @return {HTMLElement} the element
 */
function createElement(name, attrs) {
    var index = attrs ? attrs.indexOf('is') : -1;
    return index > -1
        ? rootDocument.createElement.apply(rootDocument, [name, attrs[index + 1]])
        : rootDocument.createElement(name);
}
/**
 * Handle an element.
 * @param {string} name the name
 * @param {any[]} attrs the attributes
 * @param {any[]} props the properties
 * @param {ElementOptions} opts the options
 * @return {HTMLElement} the handled element
 */
function handleElement(name, attrs, props, opts) {
    var index = nextIndex();
    var current = parentElement.childNodes.item(index);
    var found = opts.found ? opts.found : null;
    if (!found && opts.key) {
        found = utils_1.findNodeFromKey(parentElement, lastIndex(), opts.key);
    }
    if (found) {
        if (current !== found) {
            current = parentElement.insertBefore(found, current);
        }
    }
    else if (current) {
        if (name.toLowerCase() !== (current.tagName || '').toLowerCase()) {
            current = parentElement.insertBefore(createElement(name, attrs), current);
        }
    }
    else {
        current = parentElement.appendChild(createElement(name, attrs));
    }
    utils_1.updateAttributes(current, attrs);
    utils_1.updateProperties(current, props);
    if (opts.key) {
        current.dataset.fcKey = opts.key;
    }
    if (opts.content) {
        rootElement.__content__ = current;
    }
    if (!opts.skipChildren) {
        parentElement = current;
    }
    return current;
}
var funclate = {
    closeElement: closeElement,
    comment: comment,
    content: content,
    createThenUpdate: createThenUpdate,
    openElement: openElement,
    text: text,
    updateElement: updateElement,
    voidElement: voidElement
};
/**
 * Update the sub tree of the given root according to a given factory of the render function.
 * @param {RenderFactory} factory the render factory
 * @param {HTMLElement} root the root element
 * @param {Map<any>} context the context
 * @return {RenderFunction} the render function
 */
function createThenUpdate(factory, root, context) {
    if (context === void 0) { context = {}; }
    var render = factory(funclate);
    updateElement(factory(funclate), root, context);
    return render;
}
exports.createThenUpdate = createThenUpdate;
/**
 * Update the sub tree of the given root according to the given render function.
 * @param {RenderFunction} render the render function
 * @param {RootElement} root the root element
 * @param {Context} context the context
 */
function updateElement(render, root, context) {
    if (context === void 0) { context = {}; }
    var ctx = getCtx();
    rootElement = root;
    rootDocument = root.ownerDocument;
    parentElement = root;
    var lightFrag = null;
    if (!root.__visited__) {
        lightFrag = rootDocument.createDocumentFragment();
        while (root.childNodes.length > 0) {
            lightFrag.appendChild(root.removeChild(root.firstChild));
        }
    }
    render(root, context);
    cleanRemainingNodes();
    clearIndex();
    root.__visited__ = true;
    if (lightFrag && root.__content__) {
        root.__content__.appendChild(lightFrag);
    }
    restoreCtx(ctx);
}
exports.updateElement = updateElement;
/**
 * Open an element.
 *
 * @example <caption>set attributes</caption>
 * openElement('p', ['id', 'my-p' ,'class', classValue]);
 * closeElement();
 *
 * @example <caption>set properties</caption>
 * const selectValue = 'foo';
 * openElement('select', null, ['value', selectValue])
 * closeElement();
 *
 * @example <caption>set the element has a light DOM content</caption>
 * openElement('div', null, null, ['content', true])
 * closeElement();
 *
 * @example <caption>set the element's key</caption>
 * openElement('li', null, null, ['key', 'my-li'])
 * closeElement();
 *
 * @param {string} name the name of the element
 * @param {any[]} attrs the attributes of the element
 * @param {any[]} props the properties of the element
 * @param {any[]} opts the options driving the creation of the element, c.f. {@link ElementOptions}
 * @return {HTMLElement} the element
 */
function openElement(name, attrs, props, opts) {
    if (attrs === void 0) { attrs = []; }
    if (props === void 0) { props = []; }
    if (opts === void 0) { opts = []; }
    return handleElement(name, attrs, props, utils_1.fromArrayToObject(opts));
}
exports.openElement = openElement;
/**
 * Close the current element.
 */
function closeElement() {
    cleanRemainingNodes();
    clearIndex();
    parentElement = parentElement.parentElement;
}
exports.closeElement = closeElement;
/**
 * Open and close a void element.
 * @param {string} name the name of the element
 * @param {string[]} attrs the attributes of the element
 * @param {any[]} props the properties of the element
 * @param {any[]} opts the options driving the creation of the element, c.f. {@link ElementOptions}
 * @return {HTMLElement} the element
 */
function voidElement(name, attrs, props, opts) {
    if (attrs === void 0) { attrs = []; }
    if (props === void 0) { props = []; }
    if (opts === void 0) { opts = []; }
    return handleElement(name, attrs, props, utils_1.fromArrayToObject(['skipChildren', true].concat(opts)));
}
exports.voidElement = voidElement;
/**
 * To add a <code><fc-content></fc-content></code> node.
 * This node will be used to locate the entry of the light DOM structure.
 */
function content() {
    rootElement.__content__ = handleElement('fc-content', null, null, {
        found: rootElement.__content__,
        skipChildren: true
    });
}
exports.content = content;
/**
 * Create a text node.
 * @param {string} value the node value
 * @returns {Text} the text
 */
function textNodeFactory(value) {
    if (value === void 0) { value = ''; }
    return rootDocument.createTextNode(utils_1.sanitize(value));
}
/**
 * Create a text node.
 * @param {string} value the comment value
 * @returns {Comment} the comment
 */
function commentFactory(value) {
    if (value === void 0) { value = ''; }
    return rootDocument.createComment(utils_1.sanitize(value));
}
/**
 * To write a text node.
 * @param {string} value the text
 * @return {Node} the text node
 */
function text(value) {
    if (value === void 0) { value = ''; }
    return handleNode(value, rootElement.TEXT_NODE, textNodeFactory);
}
exports.text = text;
/**
 * To write a comment node.
 * @param {string} value the comment
 * @return {Node} the comment node
 */
function comment(value) {
    if (value === void 0) { value = ''; }
    return handleNode(value, rootElement.COMMENT_NODE, commentFactory);
}
exports.comment = comment;
